﻿using FLY.Thick.Blowing.IService;
using FLY.Thick.Blowing.IService.IBulkDBServicePack;
using FObjBase;
using Misc;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FLY.Thick.Blowing.Server.Model
{
    /// <summary>
    /// 提供接口，让别人获取 整理好的 当前生产，或之前生产的数据。
    /// 从LocalDB 获取数据， 或从 直接从数据获取数据
    /// 
    /// 与HistoryDB，成对。 
    /// HistoryDB 是写
    /// BulkDB 是读
    /// </summary>
    public class BulkDB : IBulkDBService
    {
        //动态的数据，不存在放在LocalDB!!!! 专门有个控件显示
        LocalDB localDB;
        DBModel dBModel;
        IShareDBService shareDB;
        /// <summary>
        /// 数据库总数量，当它变化了，肯定有新数据
        /// </summary>
        public int Count { get; set; }

        /// <summary>
        /// 开始产生数据时间
        /// </summary>
        public DateTime StartTime { get; private set; }

        /// <summary>
        /// 最后产生数据时间
        /// </summary>
        public DateTime EndTime { get; private set; }

        /// <summary>
        /// 当前生产，已经结束，已经下辊了
        /// </summary>
        public bool IsFinished { get; private set; }

        public event PropertyChangedEventHandler PropertyChanged;


        public event BulkDBTempFrameChangedEventHander TempFrameChanged;
        public BulkDB()
        {

        }

        public void SetTempFrame(int startIndex, double[] D)
        {
            TempFrameChanged?.Invoke(this, new BulkDBTempFrameChangedEventArgs() { StartIndex = startIndex, D = D });
        }

        public void Init(DBModel dBModel, IShareDBService shareDB, LocalDB localDB)
        {
            this.localDB = localDB;

            this.dBModel = dBModel;
            this.shareDB = shareDB;

            LocalDBLoad();


            Misc.BindingOperations.SetBinding(localDB, "ProfileStartTime", this, "StartTime");
            Misc.BindingOperations.SetBinding(localDB, "ProfileEndTime", this, "EndTime");
            Misc.BindingOperations.SetBinding(localDB, "IsProfileFinished", this, "IsFinished");
            
            shareDB.ScanDataChanged += ShareDB_ScanDataAdded;

        }

        private void ShareDB_ScanDataAdded(object sender, IService.IShareDBServicePack.ScanDataChangedEventArgs e)
        {
            Count++;
            //await Task.Factory.StartNew(() =>
            //{
            //    var n = dBModel.sqliteHelper.ExecuteScalar($"SELECT COUNT(*) FROM {dBModel.TbScanData.TableName}");
            //    Count = System.Convert.ToInt32(n);
            //});
        }

        /// <summary>
        /// 加载数据库中最后1次订单数据到 本地
        /// </summary>
        void LocalDBLoad()
        {
            var reponse = dBModel.sqliteHelper.ExecuteScalar($"SELECT MAX(ID) FROM {dBModel.TbProfile.TableName}");
            if (reponse is DBNull)
            {
                localDB.ProfileStartTime = DateTime.MinValue;
                localDB.ProfileEndTime = DateTime.MinValue;
                localDB.IsProfileFinished = true;
            }
            else
            {
                long maxid = System.Convert.ToInt64(reponse);
                var profile = dBModel.TbProfile.Find($"WHERE ID = {maxid}").First();
                localDB.CurrProfile = profile;
                localDB.ProfileStartTime = profile.StartTime;
                localDB.ProfileEndTime = profile.EndTime;
                localDB.IsProfileFinished = profile.IsFinished;
            }
            
            var n = dBModel.sqliteHelper.ExecuteScalar($"SELECT COUNT(*) FROM {dBModel.TbScanData.TableName}");
            Count = System.Convert.ToInt32(n);
        }
        
        /// <summary>
        /// 获取N幅数据,需要异步
        /// </summary>
        /// <param name="request"></param>
        /// <param name="AsyncDelegate"></param>
        /// <param name="AsyncContext"></param>
        public async void GetFrame(Pack_GetFrameRequest request, AsyncCBHandler AsyncDelegate, object AsyncContext)
        {
            Pack_GetFrameReponse reponse = new Pack_GetFrameReponse();
            reponse.Request = request;
            reponse.Values = new List<LC_ScanData>();
            if (request.Count > 10)//数量限制，避免死机
                request.Count = 10;

            await Task.Factory.StartNew(() =>
            {
                var db_scandatas = dBModel.TbScanData.Find($"ORDER BY ID LIMIT {request.Index},{request.Count}");
                var lc_scanDatas = AutoMapper.Mapper.Map<List<DB_ScanData>, List<LC_ScanData>>(db_scandatas);
                reponse.Values.AddRange(lc_scanDatas);
            });

            AsyncDelegate(AsyncContext, reponse);
        }

        /// <summary>
        /// 获取纵向趋势图
        /// </summary>
        /// <param name="request"></param>
        /// <param name="AsyncDelegate"></param>
        /// <param name="AsyncContext"></param>
        public async void GetTrend(Pack_GetTrendRequest request, AsyncCBHandler AsyncDelegate, object AsyncContext)
        {
            //TODO 需要异步
            Pack_GetTrendReponse reponse = new Pack_GetTrendReponse();
            reponse.Request = request;
            reponse.Values = new List<TrendValue>();
            if (request.Count > 400)//数量限制，避免死机
                request.Count = 400;
            await Task.Factory.StartNew(() =>
            {
                var db_scandatas = dBModel.TbScanData.Find($"ORDER BY ID LIMIT {request.Index},{request.Count}");
                var lc_scanDatas = AutoMapper.Mapper.Map<List<DB_ScanData>, List<LC_ScanData>>(db_scandatas);
                reponse.Values.AddRange(lc_scanDatas.Select(sd =>
                   {
                       double avg = sd.Thicks.AverageNoNull();
                       double sigma = sd.Thicks.Sigma();
                       return new TrendValue()
                       {
                           Sigma = sigma,
                           Value = avg,
                           Time = sd.Time
                       };
                   }));
            });
            AsyncDelegate(AsyncContext, reponse);
        }

        /// <summary>
        /// 完成
        /// </summary>
        public void Finish()
        {
            shareDB.FinishProfile();
        }

    }
}
